#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <sys/mman.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>
#include <linux/if_arp.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <string.h>
#include <stdlib.h>

#define PROTOCOL_ID		0x3713
#define ONLINE 0
#define OFFLINE 1
#define ERROR 2

#define BARF_DELAY 1
#define MIN_OBSERVE_DELAY 2
#define MAX_OBSERVE_DELAY 10

#define MAX_ETHS 2000

//size of eth device names
#define DEV_SIZE 30


/* Syn-3 redundancy heartbeat
 * (C)DatuX
 * Released under GPL
 */





//forks and sets realtime properties
int realtimefork()
{
	int ret=fork();
	//prevent swapping and starvation
	mlockall(MCL_CURRENT | MCL_FUTURE);
	nice(-10);
	return ret;
}

//reset serial port settings (especially echo off is important!)
void resetserial(int fh)
{
	struct termios termsettings;
	termsettings.c_lflag=ICANON;
	termsettings.c_iflag=IGNBRK;      /* input modes */
	termsettings.c_oflag=0;      /* output modes */
	termsettings.c_cflag=CBAUD|CLOCAL|CREAD|CSIZE;      /* control modes */
	memset(&termsettings.c_cc,0,sizeof(termsettings.c_cc));
	cfsetspeed(&termsettings,B38400);
	tcsetattr(fh, TCSANOW, &termsettings);
	fcntl(fh, F_SETFL, 0);
	tcflush(fh,TCIOFLUSH);
}

//Forks and barfs on the specified tty
void barftty(char tty[], char id[])
{
	if (!realtimefork())
	{
		int fh;
		char buffer[1024];
		strcpy(buffer,id);
		strcat(buffer,"\n");

		//open port
		int length=strlen(buffer);
		fh=open(tty,O_WRONLY| O_NOCTTY | O_NDELAY,0);
		if (fh<0)
			exit(ERROR);

		resetserial(fh);

		//barff
		while (1)
		{
			write(fh, buffer, length);
			sleep (BARF_DELAY);
		}
	}
}

// //Forks Observes tty for activity
// int observetty(char tty[], char id[])
// {
// 	int pid=realtimefork();
// 	if (pid)
// 	{
// 		return (pid);
// 	}
// 	else
// 	{
// 		char readbuf[1024];
// 		int fh;
//
// 		//open port
// 	 	fh=open(tty,O_RDONLY | O_NOCTTY | O_NDELAY,0);
// 		if (fh<0)
// 			exit(ERROR);
// 		resetserial(fh);
//
// 		while (1)
// 		{
// 			//receive data
// 			memset(readbuf,0,sizeof(readbuf));
// 			int readlen=read(fh, readbuf, sizeof(readbuf));
// 			if (readlen==-1)
// 				exit(ERROR);
//
// 			//is it the id?
// 			if (memcmp(readbuf, id, strlen(id))==0)
// 			{
// 				//w00t!
// 				printf("%s ",tty);
// 				fflush(NULL);
//
// 				//signal parent we have a live one:
// 				kill(getppid(), SIGUSR1);
// 			}
// 		}
// 	}
// }
//


// Converts ifindex number to interface device name (e.g. eth0)
int ifindex2dev(int s, int ifindex , char dev[])
{
	int err;

	struct ifreq ifr;
	ifr.ifr_ifindex=ifindex;
	if(err=ioctl(s, SIOCGIFNAME, &ifr))
	{
		//printf("ioctl error: %s\n", strerror(errno));
		ifr.ifr_name[0]=0;
	}
	strcpy(dev, ifr.ifr_name);
	return (err);
}

// Gets the mac adress of a network device
void getmac(int s, char dev[], char mac[])
{
	struct ifreq ifr;

	//copy device name
	strcpy(ifr.ifr_name, dev);


	//get mac
	if(ioctl(s, SIOCGIFHWADDR, &ifr))
	{
		//failed
		memset(mac,0,sizeof(struct sockaddr));
		//printf("ioctl error: %s\n", strerror(errno));
	}
	else
	{
		//ok, store mac
		memcpy(mac, ifr.ifr_hwaddr.sa_data, ETH_ALEN);

	}
}

//determine if macadress is our own
int ourmac(int s, char mac[])
{
	int ifindex;
	for (ifindex=1; ifindex<MAX_ETHS; ifindex++)
	{
		char dev[DEV_SIZE];
		char testmac[ETH_ALEN];

		ifindex2dev(s, ifindex, dev);
		getmac(s, dev, testmac);

		if (memcmp(mac, testmac, ETH_ALEN)==0)
		{
			return 1;
		}
	}
	return 0;
}


//Forks and barfs on all network interfaces.
//Uses raw ethernet, so no ip configuration needed.
void barfnet(char id[])
{
	if (!realtimefork())
	{
		//based on: http://aschauf.landshut.org/fh/linux/udp_vs_raw/ch01s03.html
		int s; /*socketdescriptor*/

		s = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
		if (s == -1)
		{
			printf("socket error\n");
			exit (ERROR);
		}

		// target adress structure, needed for sendto():
		struct sockaddr_ll socket_address;
		socket_address.sll_family   = AF_PACKET;
		socket_address.sll_protocol = htons(ETH_P_IP);
		socket_address.sll_hatype   = ARPHRD_ETHER;
		socket_address.sll_pkttype  = PACKET_BROADCAST;
		socket_address.sll_halen    = ETH_ALEN;
		//destination mac (broadcast)
		socket_address.sll_addr[0]  = 0xff;
		socket_address.sll_addr[1]  = 0xff;
		socket_address.sll_addr[2]  = 0xff;
		socket_address.sll_addr[3]  = 0xff;
		socket_address.sll_addr[4]  = 0xff;
		socket_address.sll_addr[5]  = 0xff;
		socket_address.sll_addr[6]  = 0x00;/*not used*/
		socket_address.sll_addr[7]  = 0x00;/*not used*/


		// create the ethernet packet...
		char packet[ETH_FRAME_LEN];
		memset(packet,0,sizeof(packet));

		// ..store destination mac adress in packet:
		memcpy((void*)packet, (void*)socket_address.sll_addr, ETH_ALEN);

		// ..store protocol id in packet
		struct ethhdr *eh = (struct ethhdr *)packet;
		eh->h_proto = PROTOCOL_ID;

		// ..store data in packet
		unsigned char* data = packet + 14;
		strcpy(data,id);
		int datalen=strlen(id);

		while (1)
		{
			//send packet on all interfaces:
			int ifindex;
			for (ifindex=1; ifindex<MAX_ETHS; ifindex++)
			{
				//determine device name
				char dev[DEV_SIZE];
				ifindex2dev(s, ifindex, dev);

				if (strncmp("eth",dev,3)==0)
				{

					//..store source mac adress of the destination device in packet:
					char src_mac[ETH_ALEN];
					getmac(s, dev, src_mac);
					memcpy((void*)(packet+ETH_ALEN), (void*)src_mac, ETH_ALEN);

					//select correct destination interface:
					socket_address.sll_ifindex  = ifindex;

					//send it!
					int send_result;
					send_result = sendto(
						s,
						packet,
						ETH_HLEN +datalen,
						0,
						(struct sockaddr*)&socket_address,
						sizeof(socket_address));

					if (send_result == -1 && errno!=6 && errno!= 100)
						printf("send error on interface %d: %d: %s\n", ifindex, errno, strerror(errno));
				}
			}
			sleep(BARF_DELAY);
		}
	}
}

int observenet(char id[])
{
	int pid=realtimefork();
	if (pid)
	{
		return (pid);
	}
	else
	{
		int s; /*socketdescriptor*/
		char buffer[ETH_FRAME_LEN+1000];
		char dev[DEV_SIZE];
		int length;

		//open raw socket
		s = socket(PF_PACKET, SOCK_RAW, PROTOCOL_ID);
		if (s == -1)
		{
			printf("socket error\n");
			//in case of a socket error we always pretent the node is online.
			exit (ERROR);
		}


		//flush receive buffer
		while (recvfrom(
				s,
				buffer,
				ETH_FRAME_LEN,
				MSG_DONTWAIT,
				NULL,
				NULL)> 0) ;


		while (1)
		{
			//clear buffer
			memset(buffer,0,sizeof(buffer));

			struct sockaddr_ll socket_address;
			int socket_address_size=sizeof(socket_address);

			//try to receive a frame, on any interface.
			length = recvfrom(
				s,
				buffer,
				ETH_FRAME_LEN,
				0,
				//NULL,NULL
				(struct sockaddr*)&socket_address,
				&socket_address_size
				);

			if (length<0)
				exit (ERROR);

			//is it the right length and id?
			if	(memcmp(buffer+ETH_HLEN, id, strlen(id))==0 || strcmp(id,"all")==0)
			{
				//determine if its NOT our own packet? (from one of our interfaces)
				unsigned char *mac;
				mac=buffer+ETH_ALEN;

				if (!ourmac(s, mac))
				{
					//determine device name
					ifindex2dev(s, socket_address.sll_ifindex, dev);

					//print device and source mac
					printf("%s:%02x:%02x:%02x:%02x:%02x:%02x ",
						dev,
						mac[0],
						mac[1],
						mac[2],
						mac[3],
						mac[4],
						mac[5]
					);

					//print received string, filtering out all dangerous characters
					buffer[ETH_FRAME_LEN-1]=0; //always nullterminated!
					int i=ETH_HLEN;
					while (buffer[i]!=0)
					{
						if (strchr("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890:- ",buffer[i] )!=NULL)
						{
							printf("%c", buffer[i]);
						}
						else
						{
							printf("_");
						}
						i++;
					}
					printf("\n");


					fflush(NULL);

					//signal parent we have a live one:
					kill(getppid(), SIGUSR1);
				}
			}
		}
	}
}

int final_result=OFFLINE;
//this signal is sended by our childs if it receives a heartbeat
void online_signal(int signal)
{
	final_result=ONLINE;
}

int main(int argc,const char * const *argv,const char * const *envp)
{
	if (argc==1)
	{
		printf("Usage:\n");
		printf("Barf the STRING on all interfaces    : syndog <STRING>\n");
		printf("Look for ANY identifier on all interfaces: syndog all --cat\n");
		printf("Look for identifiers _starting_ with STRING on all interfaces: syndog <STRING> --cat\n");


		exit (ERROR);
	}

	//init
	char id[1024];
	strcpy(id,argv[1]);
	//prevent starvation and swapping
	mlockall(MCL_CURRENT | MCL_FUTURE);
	nice(-10);

	//run as listening cat:
	if (argc==3)
	{
		if (strlen(id)==0)
		{
			printf("(disabled) ");
			return (OFFLINE);
		}

		//install online signal handler
		struct sigaction handler;
		handler.sa_handler=online_signal;
		handler.sa_flags=0;
		sigfillset(&handler.sa_mask);
		sigaction(SIGUSR1, &handler, NULL);

		//start processes and store pids
		int pid[10];
		int pids=0;
		//never used anymore
		// pid[pids++]=observetty("/dev/ttyS0",id);
		// pid[pids++]=observetty("/dev/ttyS1",id);
		pid[pids++]=observenet(id);

		//the program always runs for MIN_OBSERVE_DELAY seconds, but if
		//no heartbeat is received it runs for MAX_OBSERVE_DELAY
		time_t start=time(NULL);
		while(
			(time(NULL)-start) <= MIN_OBSERVE_DELAY ||
			(!final_result==ONLINE && (time(NULL)-start) < MAX_OBSERVE_DELAY)
		)
		{
			//a sleep gets interupted by signals, so we use this whileloop
			sleep(1);
		}

		//we're done, kill the childs
		int killnr;
		for(killnr=0; killnr<pids; killnr++)
		{
			kill(pid[killnr],9);
		}

		return (final_result);
	}
	//run as barfing dog:
	else
	{
		if (strlen(id))
		{
			barftty("/dev/ttyS0",id);
			barftty("/dev/ttyS1",id);
			barfnet(id);
			printf("Syndog: Barfing '%s' on all links.\n",id);
			while(wait(NULL)>0);
		}
		else
		{
			printf("Syndog: No cluster id specified. Sleeping...\n",id);
			pause();
		}
	}
}
